# Part of the jsonrpc2-zeromq-python project.
# (c) 2012 Wireless Innovation Ltd, All Rights Reserved.
# Please see the LICENSE file in the root of this project for license
# information.

from .client import *
from .server import *

from common import (RPCError, ParseError, InvalidRequest, MethodNotFound,
                    InvalidParams, InternalError, ServerError, ApplicationError,
                    JSON_RPC_VERSION, package_logger as logger)




# Part of the jsonrpc2-zeromq-python project.
# (c) 2012 Wireless Innovation Ltd, All Rights Reserved.
# Please see the LICENSE file in the root of this project for license
# information.

import threading

import zmq

from . import common


def response_from_exception(e, id_=None):
    if not hasattr(e, 'to_response'):
        e = common.ServerError(str(e))
    return e.to_response(id_)


class RPCServer(common.Endpoint, threading.Thread):

    default_socket_type = zmq.REP
    allow_methods = True
    allow_notifications = False

    should_stop = False

    def __init__(self, endpoint, context=None, timeout=1000, socket_type=None,
                 logger=None):
        super(RPCServer, self).__init__(endpoint, socket_type, timeout, context,
                                        logger=logger)
        self.socket = self.context.socket(self.socket_type)
        self.socket.bind(self.endpoint)
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)

    def stop(self):
        self.should_stop = True

    def run(self):
        self.logger.info("^_^ Server now listening on %s", self.endpoint)
        while not self.should_stop:
            self._handle_one_message()

    def _handle_one_message(self):
        req = client_id = None

        if not self.poller.poll(self.timeout):
            return

        req_parts = self.socket.recv_multipart()
        try:
            if len(req_parts) > 1:
                client_id, req = req_parts[0], req_parts[1:]
                req = ''.join(req)
            else:
                req = req_parts[0]

            try:
                req = common.json_rpc_loads(req)
            except ValueError:
                raise common.ParseError()

            if not isinstance(req, common.Request):
                raise common.InvalidRequest()

            self.logger.debug("<_< Server received {req_type} \"{method}\""
                              " on {endpoint} with params:\n{params}".format(
                                  req_type=("method call" if req.id
                                            else "notification"),
                                  method=req.method, endpoint=self.endpoint,
                                  params=common.debug_log_object_dump(
                                      req.params)))

            if (req.is_method and self.allow_methods) or \
                    (req.is_notification and self.allow_notifications):
                self._handle_method_and_response(client_id, req)

            elif req.is_method and not self.allow_methods and \
                    self.socket.socket_type in (zmq.REP, zmq.ROUTER):
                raise common.InvalidRequest(
                    "Methods not accepted by this server")

            # Cannot return an error for invalid notifications, as the spec
            # forbids it.

        except Exception as e:
            req_id = req.id if req else None
            self._send_response(client_id, req,
                                response_from_exception(e, req_id))
            if not isinstance(e, common.RPCError):
                self.logger.exception("Exception handling message in %s",
                                      self.__class__.__name__)

    def _handle_method_and_response(self, client_id, req):
        result = common.handle_request(self, 'handle_{method}_method', req)
        self._send_response(client_id, req, common.Response(result, None,
                                                            req.id))

    def _send_response(self, client_id, req, resp):
        # Notifications must not return anything
        if req and req.is_notification: return

        debug_msg_parts = [">_> Server sending"]
        debug_msg_parts.append("error" if resp.is_error else "return")
        if req:
            debug_msg_parts.append("from \"{0}\"".format(req.method))
        debug_msg_parts.append("on {0}:\n".format(self.endpoint))
        debug_msg_result = ("{indent}{0} {1}".format(
                resp.error['code'], resp.error['message'],
                        indent=common.debug_log_object_indent)
            if resp.is_error
            else common.debug_log_object_dump(resp.result))

        self.logger.debug(' '.join(debug_msg_parts) + debug_msg_result)

        self.socket.send_multipart(filter(None,
                                          [client_id,
                                           common.json_rpc_dumps(resp)]))


class RPCNotificationServer(RPCServer):

    default_socket_type = zmq.ROUTER
    allow_notifications = True


class NotificationOnlyPullServer(RPCNotificationServer):

    default_socket_type = zmq.PULL
    allow_methods = False




# Part of the jsonrpc2-zeromq-python project.
# (c) 2012 Wireless Innovation Ltd, All Rights Reserved.
# Please see the LICENSE file in the root of this project for license
# information.

import threading

import zmq

from . import common


class TimeoutError(Exception): pass


class RPCClient(common.Endpoint):

    default_socket_type = zmq.REQ
    error_code_exceptions = None
    request_method_class = common.RequestMethod

    socket = None

    def __init__(self, endpoint, context=None, timeout=5000,
                 socket_type=None, logger=None):
        super(RPCClient, self).__init__(endpoint, socket_type, timeout, context,
                                        logger)
        self.notify = NotifierProxy(self)
        self.request_poller = zmq.Poller()
        self._reconnect_socket()

    def _reconnect_socket(self):
        if self.socket:
            if self.socket in self.request_poller.sockets:
                self.request_poller.unregister(self.socket)
            self.socket.close()
        self.socket = self.context.socket(self.socket_type)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.connect(self.endpoint)
        self.request_sock = self.socket

    def request(self, request):
        self.logger.debug(">_> Client calling \"{method}\" on {endpoint} "
                       "with params:\n{params}".format(method=request.method,
                           endpoint=self.endpoint,
                           params=common.debug_log_object_dump(request.params)))

        self.request_poller.register(self.request_sock, zmq.POLLOUT)
        if not self.request_poller.poll(self.timeout):
            self.on_timeout(request)
            raise TimeoutError("Timed out while waiting to call {method} on "
                    "{endpoint}".format(method=request.method,
                        endpoint=self.endpoint))

        self.request_poller.unregister(self.request_sock)
        self.request_sock.send(common.json_rpc_dumps(request))

        if request.id is None: return # We don't get a response for notifications

        self.logger.debug("-.- Client waiting for response from {method} "
                       "on {endpoint}".format(method=request.method,
                           endpoint=self.endpoint,
                           params=common.debug_log_object_dump(request.params)))
        self.request_poller.register(self.request_sock, zmq.POLLIN)
        if not self.request_poller.poll(self.timeout):
            self.on_timeout(request)
            raise TimeoutError("Timed out while getting response to {method} on "
                    "{endpoint}".format(method=request.method,
                        endpoint=self.endpoint))

        self.request_poller.unregister(self.request_sock)
        response = common.json_rpc_loads(self.request_sock.recv())

        if request.id != response.id:
            raise ValueError("Received out-of-order response")
        if not isinstance(response, common.Response):
            raise ValueError("Received a non-response")
        if response.is_error:
            raise response.error_exception(self.error_code_exceptions)

        self.logger.debug("<_< Client received from call of \"{method}\""
                          " on {endpoint}:\n{result}".format(
                              method=request.method,
                              endpoint=self.endpoint,
                              result=common.debug_log_object_dump(response.result)))
        return response.result

    def on_timeout(self, req):
        self._reconnect_socket() # Drop outgoing message

    def get_request_method(self, method, notify=False):
        return self.request_method_class(method, client=self, notify=notify)

    def __getattr__(self, method):
        return self.get_request_method(method)


class NotifierProxy(object):

    def __init__(self, client):
        self.client = client

    def __getattr__(self, method):
        return self.client.get_request_method(method, notify=True)


class RPCNotifierClient(RPCClient):

    default_socket_type = zmq.DEALER


class NotifierOnlyPushClient(RPCClient):

    default_socket_type = zmq.PUSH


class NotificationReceiverClient(RPCNotifierClient, threading.Thread):

    on_notification = None
    should_stop = False
    poll_timeout = 1000 # milliseconds

    def __init__(self, *args, **kwargs):
        super(NotificationReceiverClient, self).__init__(*args, **kwargs)

        # We use a PAIR socket pair to communicate between the calling
        # thread and the thread (this class) receiving subscription events.
        # RPCServer thinks it's talking to the server, but now it's actually
        # talking to this thread.
        self.request_sock = self.context.socket(zmq.PAIR)
        self.request_sock_endpoint = \
                "inproc://jsonrpc2-subscription-client-%x" % id(self)
        self.request_sock.bind(self.request_sock_endpoint)

        # This is run automatically, as RPC-style blocking requests will not
        # get a response otherwise.
        self.start()

    def run(self):
        poller = zmq.Poller()
        thread_pair_sock = self.context.socket(zmq.PAIR)
        thread_pair_sock.connect(self.request_sock_endpoint)
        poller.register(thread_pair_sock, zmq.POLLIN)
        poller.register(self.socket, zmq.POLLIN)
        request_id = None

        while not self.should_stop:
            socks = dict(poller.poll(self.poll_timeout))
            if thread_pair_sock in socks and \
                    socks[thread_pair_sock] == zmq.POLLIN:
                msg = common.json_rpc_loads(thread_pair_sock.recv())
                request_id = msg.id
                self.socket.send(common.json_rpc_dumps(msg))

            if self.socket in socks and socks[self.socket] == zmq.POLLIN:
                msg_parts = self.socket.recv_multipart()
                msg = common.json_rpc_loads(msg_parts[-1])
                if msg.id and msg.id == request_id:
                    thread_pair_sock.send(common.json_rpc_dumps(msg))
                    request_id = None
                elif not msg.id:
                    self.logger.debug("<_< Client received notification "
                                      "\"{method}\" "
                                      "from subscription on {endpoint}:\n"
                                      "{result}".format(
                                          endpoint=self.endpoint,
                                          method=msg.method,
                                          result=common.debug_log_object_dump(
                                              msg.params)
                                      ))
                    common.handle_request(self, 'handle_{method}_notification',
                                          msg)

    def on_timeout(self, *args, **kwargs):
        self.stop()
        super(NotificationReceiverClient, self).on_timeout(*args, **kwargs)

    def wait_for_notifications(self):
        while self.is_alive():
            try:
                self.join(self.poll_timeout)
            except KeyboardInterrupt:
                break

    def stop(self):
        self.should_stop = True
        self.join()




# Part of the jsonrpc2-zeromq-python project.
# (c) 2012 Wireless Innovation Ltd, All Rights Reserved.
# Please see the LICENSE file in the root of this project for license
# information.

import sys
import uuid
import json
import re
import logging
import pprint

import zmq


JSON_RPC_VERSION = '2.0'

DEBUG_LOG_OBJECT_INDENT_LEN = 4


package_logger = logging.getLogger('jsonrpc2_zeromq')
package_logger.addHandler(logging.NullHandler())


debug_log_object_indent = ' ' * DEBUG_LOG_OBJECT_INDENT_LEN

def debug_log_object_dump(obj):
    return re.sub(r'(^|\n)', r'\1' + debug_log_object_indent,
                  pprint.pformat(obj, width=(80 - DEBUG_LOG_OBJECT_INDENT_LEN)))


def _json_default(o):
    if hasattr(o, 'to_dict'):
        return o.to_dict()
    elif hasattr(o, 'isoformat'):
        # Yes, UTC support only for now.
        return o.replace(microsecond=0).isoformat() + 'Z'
    else:
        raise TypeError

def _parse_rpc_message(msg):
    msg_fields = frozenset(msg.keys())
    if msg_fields.issuperset(frozenset(['jsonrpc', 'method'])):
        return Request(msg['method'], msg.get('params', None),
                id_=msg.get('id', None))
    elif msg_fields.issuperset(frozenset(['jsonrpc', 'id'])) and \
            ('result' in msg_fields or 'error' in msg_fields):
        return Response(msg.get('result', None), msg.get('error', None),
                msg['id'])
    else:
        return msg

def json_rpc_dumps(o):
    return json.dumps(o, default=_json_default)

def json_rpc_loads(s):
    return json.loads(s, object_hook=_parse_rpc_message)


_GenerateID = object()

class Request(object):

    def __init__(self, method, params, id_=_GenerateID, notify=False):
        self.method = method
        self.params = params
        if notify:
            self.id = None
        elif id_ == _GenerateID:
            self.id = str(uuid.uuid4())
        else:
            self.id = id_

    @property
    def is_method(self):
        return self.id is not None

    @property
    def is_notification(self):
        return self.id is None

    def to_dict(self):
        data = dict(jsonrpc=JSON_RPC_VERSION,
                method=self.method,
                params=self.params,
                )
        if self.id:
            data['id'] = self.id
        return data

    @property
    def method_normalised(self):
        return self.method.lower().replace('-', '_')


class Notification(Request):

    def __init__(self, method, params):
        super(Notification, self).__init__(method, params, id_=None)


class RequestMethod(object):

    def __init__(self, method, client, notify=False):
        self.method = method
        self.client = client
        self.notify = notify

    def __call__(self, *args, **kwargs):
        if args and kwargs:
            raise ValueError("Cannot be called with both positional and named "
                    "arguments")
        return self.client.request(Request(self.method, args or kwargs,
                                           notify=self.notify))


def handle_request(handler_obj, handler_attr_format, request):
    handler_id = request.method_normalised
    handler = getattr(handler_obj,
                      handler_attr_format.format(method=handler_id), None)
    if handler is None:
        raise MethodNotFound()

    try:
        if isinstance(request.params, (tuple, list)):
            result = handler(*request.params)
        elif isinstance(request.params, dict):
            result = handler(**request.params)
        else:
            raise InternalError('Parameters supplied as unexpected type')
    except TypeError as e:
        raise InvalidParams(str(e))

    return result


class Response(object):

    def __init__(self, result, error, id_):
        self.result = result
        self.error = error
        self.id = id_

    @property
    def is_error(self):
        return self.error is not None

    def error_exception(self, extra_code_mapping=None):
        cls = rpc_exception_class_for_code(self.error['code'],
                                           extra_code_mapping)
        return cls(self.error['message'], self.error.get('data', None))

    def to_dict(self):
        data = dict(jsonrpc=JSON_RPC_VERSION,
                result=self.result,
                error=self.error,
                id=self.id,
                )
        return {k: v for k, v in data.iteritems() if k == 'result' or
                                                         v is not None}


class RPCErrorMeta(type):

    def __new__(cls, name, bases, d):
        d['friendly_name'] = re.sub(r'(.)([A-Z])(?=[a-z])', r'\1 \2', name)
        out = super(RPCErrorMeta, cls).__new__(cls, name, bases, d)
        if 'error_code' in d and d['error_code'] \
                and d['error_code'] not in RPCError.class_for_error_code:
            RPCError.class_for_error_code[d['error_code']] = out
        return out


class RPCError(Exception):
    __metaclass__ = RPCErrorMeta

    error_code = None
    class_for_error_code = {}

    @property
    def error_msg(self):
        try:
            return self.args[0]
        except LookupError:
            return self.friendly_name

    @property
    def error_data(self):
        try:
            return self.args[1]
        except LookupError:
            return None

    def __str__(self):
        out = "Error {0}: {1}".format(self.error_code, self.error_msg)
        if self.error_data:
            out += "\nError data: {0!r}".format(self.error_data)
        return out

    def to_response(self, id_=None):
        return Response(None, dict(code=self.error_code, message=self.error_msg,
                                   data=self.error_data), id_)


class ParseError(RPCError):
    error_code = -32700

class InvalidRequest(RPCError):
    error_code = -32600

class MethodNotFound(RPCError):
    error_code = -32601

class InvalidParams(RPCError):
    error_code = -32602

class InternalError(RPCError):
    error_code = -32603

class ServerError(RPCError): pass
class ApplicationError(RPCError): pass


def rpc_exception_class_for_code(code, extra_code_mapping=None):
    code = code if code else -32099
    code_mapping = RPCError.class_for_error_code.copy()
    if extra_code_mapping:
        code_mapping.update(extra_code_mapping)

    cls = code_mapping.get(int(code), None)
    if cls is not None:
        return cls
    elif code >= -32099 and code <= -32000:
        return ServerError
    else:
        return ApplicationError


class Endpoint(object):

    default_socket_type = None
    error_code_exceptions = None
    logger = None

    socket = None

    def __init__(self, endpoint, socket_type, timeout, context=None,
                 logger=None):
        super(Endpoint, self).__init__()
        self.endpoint = endpoint

        if socket_type:
            self.socket_type = socket_type
        elif self.default_socket_type:
            self.socket_type = self.default_socket_type
        else:
            raise TypeError("Socket type invalid")

        self.timeout = timeout
        self.context = context or zmq.Context.instance()
        self.logger = logger if logger else package_logger

    def close(self):
        self.socket.close()




# Part of the jsonrpc2-zeromq-python project.
# (c) 2012 Wireless Innovation Ltd, All Rights Reserved.
# Please see the LICENSE file in the root of this project for license
# information.

from setuptools import setup

description = "Library for JSON-RPC 2.0 over ZeroMQ"

try:
    long_description = open(os.path.join(os.path.dirname(__file__),
                                         'README.rst')).read()
except Exception:
    long_description = description

version = "1.1.0"

setup(
    name = "jsonrpc2-zeromq",
    version = version,
    description = description,
    long_description = long_description,
    url = "https://github.com/wiltd/python-jsonrpc2-zeromq",
    download_url = "https://github.com/wiltd/"
                       "jsonrpc2-zeromq-python/tarball/v%s" % (version,),
    classifiers = [
        "Programming Language :: Python :: 2.7",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        ],
    author = "Dan Brown",
    author_email = "dan@danwb.net",
    packages = ["jsonrpc2_zeromq"],
    install_requires=[
        "distribute",
        "pyzmq<2.2",
        ],
    )

